@@ -8,7 +8,7 @@ use std::collections::HashMap;
 use std::iter::{self, FromIterator};
 use syn::{
     parse_macro_input, Attribute, AttributeArgs, Expr, Field, GenericParam, Ident, ItemStruct,
-    Lifetime, LifetimeDef, Type, TypeGenerics,
+    Lifetime, LifetimeDef, Type, TypeGenerics, TypeParamBound,
 };
 
 mod attributes;
@@ -24,7 +24,7 @@ struct StructOpts {
     /// List of attributes to apply to the generated Ref type.
     #[darling(default)]
     ref_attributes: Option<NestedMetaList>,
-    /// List of attributes to apply to the generated MutRef type.
+    /// List of attributes to apply to the generated RefMut type.
     #[darling(default)]
     ref_mut_attributes: Option<NestedMetaList>,
     /// Error type and expression to use for casting methods.
@@ -33,6 +33,9 @@ struct StructOpts {
     /// Error type and expression to use for partial getter methods.
     #[darling(default)]
     partial_getter_error: ErrorOpts,
+    /// Turn off the generation of the top-level enum that binds the variants together.
+    #[darling(default)]
+    no_enum: bool,
 }
 
 /// Field-level configuration.
@@ -146,7 +149,7 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
 
         // Drop the field-level superstruct attributes
         let mut output_field = field.clone();
-        output_field.attrs = filter_attributes(&output_field.attrs);
+        output_field.attrs = discard_superstruct_attrs(&output_field.attrs);
 
         // Add the field to the `variant_fields` map for all applicable variants.
         let field_variants = field_opts.only.as_ref().map_or_else(
@@ -203,8 +206,13 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
         output_items.push(variant_code.into());
     }
 
+    // If the `no_enum` attribute is set, stop after generating variant structs.
+    if opts.no_enum {
+        return TokenStream::from_iter(output_items);
+    }
+
     // Construct the top-level enum.
-    let top_level_attrs = filter_attributes(&item.attrs);
+    let top_level_attrs = discard_superstruct_attrs(&item.attrs);
     let enum_item = quote! {
         #(
             #top_level_attrs
@@ -229,6 +237,22 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
         0,
         GenericParam::Lifetime(LifetimeDef::new(ref_ty_lifetime.clone())),
     );
+
+    // If no lifetime bound exists for a generic param, inject one.
+    for param in ref_ty_decl_generics.params.iter_mut() {
+        if let GenericParam::Type(type_param) = param {
+            let result = type_param
+                .bounds
+                .iter()
+                .find(|bound| matches!(bound, TypeParamBound::Lifetime(_)));
+            if result.is_none() {
+                type_param
+                    .bounds
+                    .insert(0, TypeParamBound::Lifetime(ref_ty_lifetime.clone()))
+            }
+        }
+    }
+
     let (ref_impl_generics, ref_ty_generics, _) = &ref_ty_decl_generics.split_for_impl();
 
     // Prepare the attributes for the ref type.
@@ -282,30 +306,15 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
     output_items.push(ref_mut_ty.into());
 
     // Construct the main impl block.
-    let getters = fields.iter().filter(|f| f.is_common()).map(|field_data| {
-        make_field_getter(
-            type_name,
-            &variant_names,
-            &field_data.name,
-            &field_data.field.ty,
-            None,
-            &field_data.getter_opts,
-        )
-    });
+    let getters = fields
+        .iter()
+        .filter(|f| f.is_common())
+        .map(|field_data| make_field_getter(type_name, &variant_names, &field_data, None));
 
     let mut_getters = fields
         .iter()
         .filter(|f| f.is_common() && !f.getter_opts.no_mut)
-        .map(|field_data| {
-            make_mut_field_getter(
-                type_name,
-                &variant_names,
-                &field_data.name,
-                &field_data.field.ty,
-                None,
-                &field_data.getter_opts,
-            )
-        });
+        .map(|field_data| make_mut_field_getter(type_name, &variant_names, &field_data, None));
 
     let partial_getters = fields
         .iter()
@@ -319,6 +328,7 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
                 &field_variants,
                 &opts.partial_getter_error,
                 *mutability,
+                None,
             ))
         });
 
@@ -382,18 +392,35 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
         make_field_getter(
             &ref_ty_name,
             &variant_names,
-            &field_data.name,
-            &field_data.field.ty,
+            &field_data,
             Some(&ref_ty_lifetime),
-            &field_data.getter_opts,
         )
     });
 
+    let ref_partial_getters = fields
+        .iter()
+        .filter(|f| !f.is_common())
+        .flat_map(|field_data| {
+            let field_variants = field_data.only.as_ref()?;
+            Some(make_partial_getter(
+                &ref_ty_name,
+                &field_data,
+                &field_variants,
+                &opts.partial_getter_error,
+                false,
+                Some(&ref_ty_lifetime),
+            ))
+        });
+
     let ref_impl_block = quote! {
         impl #ref_impl_generics #ref_ty_name #ref_ty_generics #where_clause {
             #(
                 #ref_getters
             )*
+
+            #(
+                #ref_partial_getters
+            )*
         }
 
         // Reference types are just wrappers around references, so they can be copied!
@@ -404,7 +431,7 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
     };
     output_items.push(ref_impl_block.into());
 
-    // Construct the impl block for the *MutRef type.
+    // Construct the impl block for the *RefMut type.
     let ref_mut_getters = fields
         .iter()
         .filter(|f| f.is_common() && !f.getter_opts.no_mut)
@@ -412,18 +439,35 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
             make_mut_field_getter(
                 &ref_mut_ty_name,
                 &variant_names,
-                &field_data.name,
-                &field_data.field.ty,
+                &field_data,
                 Some(&ref_mut_ty_lifetime),
-                &field_data.getter_opts,
             )
         });
 
+    let ref_mut_partial_getters = fields
+        .iter()
+        .filter(|f| !f.is_common() && !f.partial_getter_opts.no_mut)
+        .flat_map(|field_data| {
+            let field_variants = field_data.only.as_ref()?;
+            Some(make_partial_getter(
+                &ref_mut_ty_name,
+                &field_data,
+                &field_variants,
+                &opts.partial_getter_error,
+                true,
+                Some(&ref_mut_ty_lifetime),
+            ))
+        });
+
     let ref_mut_impl_block = quote! {
         impl #ref_mut_impl_generics #ref_mut_ty_name #ref_mut_ty_generics #where_clause {
             #(
                 #ref_mut_getters
             )*
+
+            #(
+                #ref_mut_partial_getters
+            )*
         }
     };
     output_items.push(ref_mut_impl_block.into());
@@ -435,27 +479,32 @@ pub fn superstruct(args: TokenStream, input: TokenStream) -> TokenStream {
 fn make_field_getter(
     type_name: &Ident,
     variant_names: &[Ident],
-    field_name: &Ident,
-    field_type: &Type,
+    field_data: &FieldData,
     lifetime: Option<&Lifetime>,
-    getter_opts: &GetterOpts,
 ) -> proc_macro2::TokenStream {
+    let field_name = &field_data.name;
+    let field_type = &field_data.field.ty;
+    let getter_opts = &field_data.getter_opts;
+
     let fn_name = getter_opts.rename.as_ref().unwrap_or(field_name);
     let return_type = if getter_opts.copy {
         quote! { #field_type }
     } else {
-        if let Some(lifetime) = lifetime {
-            quote! { &#lifetime #field_type}
-        } else {
-            quote! { &#field_type}
-        }
+        quote! { &#lifetime #field_type}
     };
     let return_expr = if getter_opts.copy {
         quote! { inner.#field_name }
     } else {
         quote! { &inner.#field_name }
     };
+
+    // Pass-through `cfg` attributes as they affect the existence of this field.
+    let cfg_attrs = get_cfg_attrs(&field_data.field.attrs);
+
     quote! {
+        #(
+            #cfg_attrs
+        )*
         pub fn #fn_name(&self) -> #return_type {
             match self {
                 #(
@@ -472,24 +521,25 @@ fn make_field_getter(
 fn make_mut_field_getter(
     type_name: &Ident,
     variant_names: &[Ident],
-    field_name: &Ident,
-    field_type: &Type,
+    field_data: &FieldData,
     lifetime: Option<&Lifetime>,
-    getter_opts: &GetterOpts,
 ) -> proc_macro2::TokenStream {
+    let field_name = &field_data.name;
+    let field_type = &field_data.field.ty;
+    let getter_opts = &field_data.getter_opts;
+
     let fn_name = format_ident!("{}_mut", getter_opts.rename.as_ref().unwrap_or(field_name));
-    let return_type = if let Some(lifetime) = lifetime {
-        quote! { &#lifetime mut #field_type}
-    } else {
-        quote! { &mut #field_type}
-    };
-    let param = if let Some(lifetime) = lifetime {
-        quote! { &#lifetime mut self}
-    } else {
-        quote! { &mut self}
-    };
+    let return_type = quote! { &#lifetime mut #field_type };
+    let param = make_self_arg(true, lifetime);
     let return_expr = quote! { &mut inner.#field_name };
+
+    // Pass-through `cfg` attributes as they affect the existence of this field.
+    let cfg_attrs = get_cfg_attrs(&field_data.field.attrs);
+
     quote! {
+        #(
+            #cfg_attrs
+        )*
         pub fn #fn_name(#param) -> #return_type {
             match self {
                 #(
@@ -502,22 +552,29 @@ fn make_mut_field_getter(
     }
 }
 
-fn make_self_arg(mutable: bool) -> proc_macro2::TokenStream {
+fn make_self_arg(mutable: bool, lifetime: Option<&Lifetime>) -> proc_macro2::TokenStream {
     if mutable {
-        quote! { &mut self }
+        quote! { &#lifetime mut self }
     } else {
+        // Ignore the lifetime for immutable references. This allows `&Ref<'a>` to be de-referenced
+        // to an inner pointer with lifetime `'a`.
         quote! { &self }
     }
 }
 
-fn make_type_ref(ty: &Type, mutable: bool, copy: bool) -> proc_macro2::TokenStream {
+fn make_type_ref(
+    ty: &Type,
+    mutable: bool,
+    copy: bool,
+    lifetime: Option<&Lifetime>,
+) -> proc_macro2::TokenStream {
     // XXX: bit hacky, ignore `copy` if `mutable` is set
     if mutable {
-        quote! { &mut #ty }
+        quote! { &#lifetime mut #ty }
     } else if copy {
         quote! { #ty }
     } else {
-        quote! { &#ty }
+        quote! { &#lifetime #ty }
     }
 }
 
@@ -528,6 +585,7 @@ fn make_partial_getter(
     field_variants: &[Ident],
     error_opts: &ErrorOpts,
     mutable: bool,
+    lifetime: Option<&Lifetime>,
 ) -> proc_macro2::TokenStream {
     let field_name = &field_data.name;
     let renamed_field = field_data
@@ -541,8 +599,8 @@ fn make_partial_getter(
         renamed_field.clone()
     };
     let copy = field_data.partial_getter_opts.copy;
-    let self_arg = make_self_arg(mutable);
-    let ret_ty = make_type_ref(&field_data.field.ty, mutable, copy);
+    let self_arg = make_self_arg(mutable, lifetime);
+    let ret_ty = make_type_ref(&field_data.field.ty, mutable, copy, lifetime);
     let ret_expr = if mutable {
         quote! { &mut inner.#field_name }
     } else if copy {
@@ -551,7 +609,14 @@ fn make_partial_getter(
         quote! { &inner.#field_name }
     };
     let (res_ret_ty, err_expr) = error_opts.build_result_type(&ret_ty);
+
+    // Pass-through `cfg` attributes as they affect the existence of this field.
+    let cfg_attrs = get_cfg_attrs(&field_data.field.attrs);
+
     quote! {
+        #(
+            #cfg_attrs
+        )*
         pub fn #fn_name(#self_arg) -> #res_ret_ty {
             match self {
                 #(
@@ -600,7 +665,7 @@ fn make_as_variant_method(
 }
 
 /// Keep all non-superstruct-related attributes from an array.
-fn filter_attributes(attrs: &[Attribute]) -> Vec<Attribute> {
+fn discard_superstruct_attrs(attrs: &[Attribute]) -> Vec<Attribute> {
     attrs
         .iter()
         .filter(|attr| !is_superstruct_attr(attr))
@@ -608,9 +673,23 @@ fn filter_attributes(attrs: &[Attribute]) -> Vec<Attribute> {
         .collect()
 }
 
+/// Keep only `cfg` attributes from an array.
+fn get_cfg_attrs(attrs: &[Attribute]) -> Vec<Attribute> {
+    attrs
+        .iter()
+        .filter(|attr| is_attr_with_ident(attr, "cfg"))
+        .cloned()
+        .collect()
+}
+
 /// Predicate for determining whether an attribute is a `superstruct` attribute.
 fn is_superstruct_attr(attr: &Attribute) -> bool {
+    is_attr_with_ident(attr, "superstruct")
+}
+
+/// Predicate for determining whether an attribute has the given `ident` as its path.
+fn is_attr_with_ident(attr: &Attribute, ident: &str) -> bool {
     attr.path
         .get_ident()
-        .map_or(false, |ident| ident.to_string() == "superstruct")
+        .map_or(false, |attr_ident| attr_ident.to_string() == ident)
 }
